#!/usr/bin/env python
# Copyright (c) 2003-2014  Pavel Rychly, Vojtech Kovar, Milos Jakubicek, Milos Husak, Vit Baisa

import manatee
import os, re, time, sys
from butils import set_proctitle
from conccache import add_to_map, del_from_map, load_map, get_cached_conc,\
                      get_existing_conc, get_existing_conc_sizes
from pyconc import PyConc

def tokens2strclass (tokens):
    return [{'str': tokens[i], 'class': tokens[i+1].strip ('{}')}
            for i in range(0, len(tokens), 2)]


def printkwic (conc, froml=0, tol=5, leftctx='15#', rightctx='15#',
               attrs='word', refs='#', maxcontext=0):
    def strip_tags (tokens):
        return ''.join ([tokens[i] for i in range(0, len(tokens), 2)])
    kl = manatee.KWICLines (conc.corp(), conc.RS (True), leftctx, rightctx,
                            attrs, 'word', 'p', refs, maxcontext)
    while kl.nextline ():
        print '%s %s <%s> %s' % (kl.get_refs(), strip_tags (kl.get_left()),
                                 strip_tags (kl.get_kwic()),
                                 strip_tags (kl.get_right()))


def pos_ctxs (min_hitlen, max_hitlen, max_ctx=3):
    ctxs = [{'n': _('%iL') % -c, 'ctx':'%i<0' % c} for c in range (-max_ctx, 0)]
    if max_hitlen == 1:
        ctxs.append ({'n': _('Node'), 'ctx': '0~0>0'})
    else:
        ctxs.extend ([{'n':'Node %i' % c, 'ctx':'%i<0' % c}
                      for c in range (1,max_hitlen+1)])
    ctxs.extend ([{'n': _('%iR') % c, 'ctx':'%i>0' % c}
                  for c in range (1, max_ctx+1)])
    return ctxs


def kwicpage (conc, fromp=1, leftctx='40#', rightctx='40#', attrs='word',
              ctxattrs='word', refs='#', structs='p', pagesize=20,
              labelmap={}, righttoleft=False, alignlist=[], copy_icon=0,
              tbl_template='none', hidenone=0):
    refs = refs.replace('.MAP_OUP', '') # to be removed ...
    try:
        fromp = int(fromp)
        if fromp < 1:
            fromp = 1
    except:
        fromp = 1
    corps_with_colls = manatee.StrVector()
    conc.get_aligned(corps_with_colls)
    if not conc.orig_corp.get_conffile() in corps_with_colls:
        kwcl = kwiclines (conc, (fromp -1) * pagesize, fromp * pagesize,
                          '0', '0', 'word', '', refs, structs,
                          labelmap, righttoleft)
    else:
        kwcl = kwiclines (conc, (fromp -1) * pagesize, fromp * pagesize,
                          leftctx, rightctx, attrs, ctxattrs, refs, structs,
                          labelmap, righttoleft)
    out = {'Lines': kwcl}
    add_aligns(out, conc,(fromp -1) * pagesize, fromp * pagesize,
               leftctx, rightctx, attrs, ctxattrs, refs, structs,
               labelmap, righttoleft, alignlist)
    if copy_icon:
        from tbl_settings import tbl_refs, tbl_structs
        sen_refs = tbl_refs.get(tbl_template, '') + ',#'
        sen_refs = sen_refs.replace('.MAP_OUP', '') # to be removed ...
        sen_structs = tbl_structs.get(tbl_template, '') or 'g'
        sen_lines = kwiclines(conc, (fromp -1) * pagesize, fromp * pagesize,
                             '-1:s', '1:s', refs=sen_refs, structs=sen_structs)
        for old, new in zip(out['Lines'], sen_lines):
            old['Sen_Left'] = new['Left']
            old['Sen_Right'] = new['Right']
            old['Tbl_refs'] = new['Tbl_refs']
    if labelmap:
        out['GroupNumbers'] = format_labelmap (labelmap)
    if fromp > 1:
        out['prevlink'] = 'fromp=%i' % (fromp - 1)
        out['firstlink'] = 'fromp=1'
    if conc.size() > pagesize:
        out['fromp'] = fromp
        out['numofpages'] = numofpages = (conc.size()-1)/pagesize + 1
        if fromp < numofpages:
            out['nextlink'] = 'fromp=%i' % (fromp +1)
            out['lastlink'] = 'fromp=%i' % numofpages
    out['concsize'] = conc.size()
    if hidenone:
        for line in out['Lines']:
            for part in ('Kwic', 'Left', 'Right'):
                for item in line[part]:
                    item['str'] = item['str'].replace('===NONE===', '')
                line['ref'] = line['ref'].replace('===NONE===', '')
    return out

def add_aligns(result, conc, fromline, toline, leftctx='40#', rightctx='40#',
               attrs='word', ctxattrs='word', refs='#', structs='p',
               labelmap={}, righttoleft=False, alignlist=[]):
    if not alignlist: return
    al_lctx = leftctx.endswith('#') and '0' or leftctx
    al_rctx = rightctx.endswith('#') and '0' or rightctx
    al_lines = []
    corps_with_colls = manatee.StrVector()
    conc.get_aligned(corps_with_colls)
    result['CollCorps'] = list(corps_with_colls)
    result['Par_conc_corpnames'] = [{'n': c.get_conffile(),
                                     'label': c.get_conf('NAME')
                                              or c.get_conffile()}
                                    for c in [conc.orig_corp]+alignlist]
    for al_corp in alignlist:
        al_corpname = al_corp.get_conffile()
        if al_corpname in corps_with_colls:
            conc.switch_aligned (al_corp.get_conffile())
            al_lines.append (kwiclines (conc, fromline, toline, leftctx,
                                        rightctx, attrs, ctxattrs, refs,
                                        structs, labelmap, righttoleft))
        else:
            conc.switch_aligned(conc.orig_corp.get_conffile())
            conc.add_aligned(al_corp.get_conffile())
            conc.switch_aligned (al_corp.get_conffile())
            al_lines.append (kwiclines (conc, fromline, toline, al_lctx,
                                        al_rctx, 'word', '', refs, structs,
                                        labelmap, righttoleft))
    aligns = zip(*al_lines)
    for i, line in enumerate(result['Lines']):
        line['Align'] = aligns[i]
    
def kwiclines (conc, fromline, toline, leftctx='40#', rightctx='40#',
               attrs='word', ctxattrs='word', refs='#', structs='p',
               labelmap={}, righttoleft=False):
    def non1hitlen (hitlen):
        if hitlen == 1: return ''
        else: return ';hitlen=%i' % hitlen

    lines = []
    if righttoleft:
        rightlabel, leftlabel = 'Left', 'Right'
        structs += ',ltr'
        #from unicodedata import bidirectional
        def isengword (strclass):
            #return bidirectional(word[0]) in ('L', 'LRE', 'LRO')
            return 'ltr' in strclass['class'].split()
    else:
        leftlabel, rightlabel = 'Left', 'Right'

    kl = manatee.KWICLines (conc.corp(), conc.RS (True, fromline, toline),
                            leftctx, rightctx, attrs, ctxattrs, structs, refs)
    labelmap = labelmap.copy()
    labelmap[0] = '_'
    maxleftsize = 0
    maxrightsize = 0
    while kl.nextline():
        linegroup = kl.get_linegroup() or 0
        linegroup = labelmap.get(linegroup, '_')
        leftwords = tokens2strclass (kl.get_left())
        rightwords = tokens2strclass (kl.get_right())
        kwicwords = tokens2strclass (kl.get_kwic())
        if righttoleft:
            # change order for "English" context of "English" keywords
            if isengword(kwicwords[0]):
                # preceding words
                nprev = len(leftwords) -1
                while nprev >= 0 and isengword (leftwords[nprev]):
                    nprev -= 1
                if nprev == -1:
                    # move whole context
                    moveleft = leftwords
                    leftwords = []
                else:
                    moveleft = leftwords[nprev+1:]
                    del leftwords[nprev+1:]

                # following words
                nfollow = 0
                while (nfollow < len(rightwords)
                       and isengword (rightwords[nfollow])):
                    nfollow += 1
                moveright = rightwords[:nfollow]
                del rightwords[:nfollow]
                
                leftwords = leftwords + moveright
                rightwords = moveleft + rightwords

        leftsize = 0
        for w in leftwords:
            if not w['class'] == 'strc':
                leftsize += len(w['str']) + 1
        if leftsize > maxleftsize:
            maxleftsize = leftsize

        rightsize = 0
        for w in rightwords:
            if not w['class'] == 'strc':
                rightsize += len(w['str']) + 1
        if rightsize > maxrightsize:
            maxrightsize = rightsize
                
        lines.append ({'toknum': kl.get_pos(),
                       'hitlen': non1hitlen (kl.get_kwiclen()),
                       'ref': kl.get_refs(),
                       'Tbl_refs': list(kl.get_ref_list()),
                       leftlabel: leftwords,
                       'Kwic': kwicwords,
                       rightlabel: rightwords,
                       'linegroup': linegroup,
                       'leftsize': leftsize,
                       'rightsize': rightsize,
                       })
    for l in lines:
        l['leftspace'] = ' ' * (maxleftsize - l['leftsize'])
        l['rightspace'] = ' ' * (maxrightsize - l['rightsize'])
    return lines


def strkwiclines (conc, fromline, toline=None, leftctx='40#', rightctx='40#'):
    def tokens2str (tokens):
        return ''.join ([tokens[i] for i in xrange(0, len(tokens), 2)])
    toline = toline or fromline +1
    kl = manatee.KWICLines (conc.corp(), conc.RS (True, fromline, toline),
                            leftctx, rightctx, 'word', 'word','','')
    return [{'left': tokens2str (kl.get_left()),
             'kwic': tokens2str (kl.get_kwic()),
             'right': tokens2str (kl.get_right())}
            for line in xrange (fromline, toline) if kl.nextline ()]


def get_sort_idx (conc, q=[], pagesize=20, enc='latin1'):
    crit = ''
    for qq in q:
        if qq.startswith('s') and not qq.startswith('s*'): crit = qq[1:]
    if not crit: return []
    vals = manatee.StrVector(); idx = manatee.IntVector()
    if '.' in crit.split('/')[0]: just_letters = False
    else: just_letters = True
    conc.sort_idx(crit, vals, idx, just_letters)
    out = [(v, pos/pagesize + 1) for v, pos in zip(vals, idx)]
    if just_letters:
        result = []; keys = []
        for v, p in out:
            if not v[0] in keys:
                result.append((v[0], p)); keys.append(v[0])
        out = result
    return [{'page': p, 'label': v} for v, p in out]

def get_conc_sizes (corp, q=[], _cache_dir="cache", server_port=None):
    return get_existing_conc_sizes (corp, q, _cache_dir, server_port)

def compute_conc (corp, q, _cache_dir, subchash, samplesize, fullsize, pid_dir,
                  save):
    q = tuple (q)
    if q[0][0] == "R": # online sample
        if fullsize == -1: # need to compute original conc first
            q_copy = list(q)
            q_copy[0] = q[0][1:]
            q_copy = tuple(q_copy)
            conc = get_sync_conc (corp, q_copy, save, _cache_dir, pid_dir,
                                  subchash, samplesize, fullsize)
            fullsize = conc.fullsize()
        return PyConc (corp, q[0][1], q[0][2:], samplesize, fullsize)
    else:
        return PyConc (corp, q[0][0], q[0][1:], samplesize)

def get_sync_conc (corp, q, save, _cache_dir, pid_dir, subchash, samplesize,
                   fullsize):
    conc = None
    if save:
        cachefile, pidfile, server = add_to_map (_cache_dir, pid_dir,
                                                 subchash, q[:1], 0)
        if type(server) == int: # computation got started meanwhile
            conc = get_existing_conc (corp, subchash, q[:1], _cache_dir,
                                      pid_dir, -1, server)
    if not conc:
        try:
            conc = compute_conc (corp, q, _cache_dir, subchash, samplesize,
                                 fullsize, pid_dir, save)
            conc.sync() # wait for the computation to finish
        except:
            del_from_map (_cache_dir, subchash, q)
            raise
    if save:
        conc.save (cachefile)
        # update size in map file
        add_to_map (_cache_dir, pid_dir, subchash, q[:1], conc.size())
        os.remove (pidfile)
    return conc

def get_async_conc (corp, q, save, _cache_dir, pid_dir, subchash, samplesize,
                    fullsize, minsize):
    r, w = os.pipe()
    r, w = os.fdopen(r,'r'), os.fdopen(w,'w')
    if os.fork() == 0: # child
        r.close() # child writes
        title = "conc;%s;%s;%s;" % (corp.get_conffile(), q[0][0], q[0][1:])
        set_proctitle(title.encode("utf-8"))
        # close stdin/stdout/stderr so that the webserver closes
        # connection to client when parent ends, comment for debugging
        os.close(0); os.close(1); os.close(2)
        try:
            cachefile, pidfile, server = add_to_map(_cache_dir, pid_dir,
                                                    subchash, q, -1)
            if type(server) == int:
                # conc got started meanwhile by another process
                w.write (server)
                w.close()
                os._exit(0)
            sys.stderr = server.logfile
            conc = compute_conc (corp, q, _cache_dir, subchash,
                                 samplesize, fullsize, pid_dir, save)
            w.write (str(server.server_port))
            w.close()
            server.timeout = 5
            server.conc = conc
            while not conc.finished():
                server.handle_request()
            conc.save (cachefile)
            # update size in map file
            add_to_map (_cache_dir, pid_dir, subchash, q, conc.size())
            os.remove (pidfile)
            os._exit(0)
        except Exception, e:
            import traceback, cPickle
            if not w.closed:
                stack = traceback.format_exception(*sys.exc_info())
                e = (e, stack)
                w.write (cPickle.dumps(e))
                w.close()
            traceback.print_exc (None, open(pidfile, "a"))
            os._exit(0)
    else: # parent
        w.close() # parent reads
        server_port = r.read()
        try:
            server_port = int (server_port)
        except:
            import cPickle
            e = cPickle.loads (server_port)
            args = list(e[0].args)
            args.append("".join(e[1]))
            e[0].args = tuple(args)
            raise e[0]
        r.close()
        conc = get_existing_conc (corp, subchash, q, _cache_dir, pid_dir,
                                  minsize, server_port)
        conc.port = server_port
        return conc

def get_conc (corp, minsize=None, q=[], fromp=0, pagesize=0, async=0, save=0, \
              _cache_dir='cache', samplesize=0, debug=False):
    if not q:
        return None
    q = tuple (q)
    if not minsize:
        if len(q) > 1 or not async: # subsequent concordance processing
                                    # by its methods needs whole concordance
            minsize = -1
        else:
            minsize = fromp * pagesize
    _cache_dir = _cache_dir + '/' + corp.corpname + '/'
    pid_dir = _cache_dir + "/run/"
    subchash = getattr(corp, 'subchash', None)
    conc = None
    fullsize = -1

    # try to locate concordance in cache
    if save:
        toprocess, conc = get_cached_conc(corp, subchash, q, _cache_dir,
                                          pid_dir, minsize)
        if toprocess == len(q):
            save = 0
        if not conc and q[0][0] == "R": # online sample
            q_copy = list(q)
            q_copy[0] = q[0][1:]
            t, c = get_cached_conc (corp, subchash, q_copy, _cache_dir,
                                    pid_dir, -1)
            if c:
                fullsize = c.fullsize()
    else:
        async = 0

    # cache miss or not used
    if not conc:
        toprocess = 1
        if async and len(q) == 1: # asynchronous processing
            conc = get_async_conc (corp, q, save, _cache_dir, pid_dir,
                                   subchash, samplesize, fullsize, minsize)
        else: # synchronous processing
            conc = get_sync_conc (corp, q, save, _cache_dir, pid_dir,
                                  subchash, samplesize, fullsize)

    # process subsequent concordance actions
    for act in range(toprocess, len(q)):
        command = q[act][0]
        getattr (conc, 'command_' + command) (q[act][1:])
        if command in 'gasL':# user specific/volatile actions, cannot save
            save = 0
        if save:
            cachefile, pidfile, server = add_to_map (_cache_dir, pid_dir,
                                                     subchash, q[:act + 1],
                                                     conc.size())
            if type(server) != int: # nobody started the computation yet
                conc.save (cachefile)
                os.remove (pidfile)
    return conc


def get_conc_desc (q=[], _cache_dir='cache', corpname='', subchash=None):
    desctext = {'q': _('Query'),
                'a': _('Query'),
                'L': _('Label filter'),
                'R': _('Query'),
                'r': _('Random sample'),
                's': _('Sort'),
                'f': _('Shuffle'),
                'D': _('Subparts filter'),
                'F': _('Filter all but first hit in document'),
                'n': _('Negative filter'),
                'N': _('Negative filter (excluding KWIC)'),
                'p': _('Positive filter'),
                'P': _('Positive filter (excluding KWIC)'),
                'w': _('Word sketch item'),
                't': _('Term'),
                'e': _('GDEX'),
                'x': _('Switch KWIC'),
                'g': _('Sort labels'),
                }
    forms = {'q': ('first_form', 'cql'),
             'a': ('first_form', 'cql'),
             'r': ('reduce_form', 'rlines'),
             }
    desc = []
    saved = load_map (_cache_dir + '/' + corpname + '/')
    q = tuple (q)

    for i in range (len(q)):
        size = saved.get ((subchash, q[:i+1]), ('',''))[1]
        opid = q[i][0]
        if opid == 'L':
            args = str(abs(int(q[i].split()[-1])))
        else:
            args = q[i][1:]
        url1p = [('q', qi) for qi in q[:i]]
        url2p = [('q', qi) for qi in q[:i+1]]
        op = desctext.get (opid)
        formname = forms.get(opid, ('',''))
        if formname[1]:
            if formname[1] == 'cql':
                if opid == 'a': default_attr, url1args = args.split(',', 1)
                else: default_attr, url1args = 'word', args
                url1p.append (('queryselector', 'cqlrow'))
                url1p.append (('default_attr', default_attr))
            else:
                url1args = args
            url1p.append ((formname[1], url1args))

        if opid == 's' and args[0] != '*' and i > 0 and len (args.split()) > 2:
            op = 'Multilevel Sort'
        if op:
            if not formname[0]: url1p = ''
            desc.append ((op, args, url1p, url2p, size, formname[0]))
    return desc

def get_conc_labelmap(infopath):
    labels = {}
    try:
        for line in open(infopath):
            try:
                n, lab = line.strip().split('\t')
            except ValueError:
                continue
            labels[int(n)] = lab
        if not labels:
            raise Exception(_('Empty .info file'))
    except IOError, err:
        raise Exception(_('Error while reading .info file'))
    return labels

number_re = re.compile ('[0-9]+$')

def lngrp_sortcrit (lab, separator='.'):
    def num2sort (n):
        if number_re.match (n):
            return ('n', int(n))
        else:
            if n == 'u':
                return ('y', n)
            if n == 'x':
                return ('z', n)
            return ('c', n)
    if not lab:
        return [('x','x')]
    return map (num2sort, lab.split (separator, 3))

def lngrp_sortstr (lab, separator='.'):
    f = {'n': 'n%03g', 'c': 'c%s', 'x': '%s', 'y': 'y%s', 'z': 'z%s'}
    return '|'.join([f[c] % s for c,s in lngrp_sortcrit(lab, separator)])


def format_labelmap (labelmap, separator='.'):
    matrix = {}
    for n,lab in labelmap.items():
        if lab:
            pref = lab.split (separator)[0]
            matrix.setdefault (pref, []).append ((lngrp_sortcrit(lab), lab, n))
    prefixes = [(lngrp_sortcrit(p), p) for p in matrix.keys()]
    prefixes.sort()
    lines = []
    for s, pref in prefixes:
        line = matrix[pref]
        line.sort()
        lines.append({
            'Items': [{'n': n, 'lab': lab} for (s, lab, n) in line],
            'Pref': pref
        })
    return lines


def get_stored_conc (corp, concname, conc_dir):
    basecorpname = corp.corpname.split(':')[0]
    conc_dir = os.path.join (conc_dir, basecorpname)
    if not os.path.isdir (conc_dir):
        os.makedirs (conc_dir)
    cpath = os.path.join (conc_dir, concname)
    conc = PyConc (corp, 'l', cpath + '.conc')
    conc.labelmap = get_conc_labelmap (cpath + '.info')
    return conc


def get_full_ref (corp, pos):
    data = {}
    refs = [(n == '#' and ('#', str(pos)) or
             (n, corp.get_attr(n).pos2str (pos)))
            for n in corp.get_conf ('FULLREF').split(',')]
    data['Refs'] = [{'name': n == '#' and _('Token number')
                             or corp.get_conf (n+'.LABEL') or n,
                     'val': v}
                    for n,v in refs]
    for n,v in refs:
        data [n.replace('.','_')] = v
    return data


def get_detail_context (corp, pos, hitlen=1,
                        detail_left_ctx=40, detail_right_ctx=40,
                        addattrs=[], structs='', detail_ctx_incr=60):
    data = {}
    wrapdetail = corp.get_conf ('WRAPDETAIL')
    if wrapdetail:
        data['wrapdetail'] = '<%s>' % wrapdetail
        if not wrapdetail in structs.split(','): data['deletewrap'] = True
        structs = wrapdetail + ',' + structs
    else:
        data['wrapdetail'] = ''
    try:
        maxdetail = int (corp.get_conf ('MAXDETAIL'))
    except:
        maxdetail = 0
    if maxdetail:
        if detail_left_ctx > maxdetail:
            detail_left_ctx = maxdetail
        if detail_right_ctx > maxdetail:
            detail_right_ctx = maxdetail
    if detail_left_ctx > pos:
        detail_left_ctx = pos
    attrs = ','.join(['word'] + addattrs)
    cr = manatee.CorpRegion(corp, attrs, structs)
    region_left = tokens2strclass (cr.region (pos - detail_left_ctx, pos))
    region_kwic = tokens2strclass (cr.region (pos, pos + hitlen))
    region_right = tokens2strclass (cr.region(pos + hitlen,
                                              pos + hitlen + detail_right_ctx))
    for seg in region_left + region_kwic + region_right:
        seg['str'] = seg['str'].replace('===NONE===', '')
    for seg in region_kwic:
        if not seg['class']: seg['class'] = 'coll'
    data['content'] = region_left + region_kwic + region_right
    refbase = 'pos=%i;' % pos
    if hitlen != 1: refbase += 'hitlen=%i;' % hitlen
    data['leftlink'] = refbase + ('detail_left_ctx=%i;detail_right_ctx=%i'
                                  % (detail_left_ctx + detail_ctx_incr,
                                     detail_right_ctx))
    data['rightlink'] = refbase + ('detail_left_ctx=%i;detail_right_ctx=%i'
                                   % (detail_left_ctx,
                                      detail_right_ctx + detail_ctx_incr))
    if corp.get_conf('RIGHTTOLEFT') == '1': data['righttoleft'] = True
    data['pos'] = pos
    return data


def fcs_search(corp, fcs_query, max_rec, start):
    "aux function for federated content search: operation=searchRetrieve"
    if not fcs_query:
        raise Exception(7, 'fcs_query', 'Mandatory parameter not supplied')
    query = fcs_query.replace('+', ' ') # convert URL spaces
    exact_match = False # attr=".*value.*"
    if 'exact' in query.lower() and not '=' in query: # lemma EXACT "dog"
        pos = query.lower().index('exact') # first occurence of EXACT
        query = query[:pos] + '=' + query[pos+5:] # 1st exact > =
        exact_match = True
    rq = '' # query for manatee
    try: # parse query
        if '=' in query: # lemma=word | lemma="word" | lemma="w1 w2" | word=""
            attr, term = query.split('=')
            attr = attr.strip()
            term = term.strip()
        else: # "w1 w2" | "word" | word
            if 'lemma' in corp.get_conf('ATTRLIST').split(','):
                attr = 'lemma'
            else:
                attr = 'word'
            term = query.strip()
        if '"' in attr:
            raise Exception
        if '"' in term: # "word" | "word1 word2" | "" | "it is \"good\""
            if term[0] != '"' or term[-1] != '"': # check q. marks
                raise Exception
            term = term[1:-1].strip() # remove quotation marks
            if ' ' in term: # multi-word term
                if exact_match:
                    rq = ' '.join(['[%s="%s"]' % (attr, t)
                            for t in term.split()])
                else:
                    rq = ' '.join(['[%s=".*%s.*"]' % (attr, t)
                            for t in term.split()])
            elif term.strip() == '': # ""
                raise Exception # empty term
            else: # one-word term
                if exact_match:
                    rq = '[%s="%s"]' % (attr, term)
                else:
                    rq = '[%s=".*%s.*"]' % (attr, term)
        else: # must be single-word term
            if ' ' in term:
                raise Exception
            if exact_match: # build query
                rq = '[%s="%s"]' % (attr, term)
            else:
                rq = '[%s=".*%s.*"]' % (attr, term)
    except: # there was a problem when parsing
        raise Exception(10, query, 'Query syntax error')
    if not attr in corp.get_conf('ATTRLIST'):
        raise Exception(16, attr, 'Unsupported index')
    try: # try to get concordance
        conc = get_conc(corp, q=['q' + rq])
    except Exception, e:
        raise Exception(10, repr(e), 'Query syntax error')
    page = kwicpage(conc) # convert concordance
    if len(page['Lines']) < start:
        raise Exception(61, 'startRecord', 'First record position out of range')
    return [(kwicline['Left'][0]['str'], kwicline['Kwic'][0]['str'],
            kwicline['Right'][0]['str'], kwicline['ref']) 
            for kwicline in page['Lines']][start:][:max_rec]


def fcs_scan(corpname, scan_query, max_ter, start):
    "aux function for federated content search: operation=scan"
    if not scan_query:
        raise Exception(7, 'scan_query', 'Mandatory parameter not supplied')
    query = scan_query.replace('+', ' ') # convert URL spaces
    exact_match = False
    if 'exact' in query.lower() and not '=' in query: # lemma ExacT "dog"
        pos = query.lower().index('exact') # first occurence of EXACT
        query = query[:pos] + '=' + query[pos+5:] # 1st exact > =
        exact_match = True
    corp = manatee.Corpus(corpname)
    attrs = corp.get_conf('ATTRLIST').split(',') # list of available attrs
    try:
        if '=' in query:
            attr, value = query.split('=')
            attr = attr.strip()
            value = value.strip()
        else: # must be in format attr = value
            raise Exception
        if '"' in attr:
            raise Exception
        if '"' in value:
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1].strip()
            else:
                raise Exception
    except Exception, e:
        raise Exception(10, scan_query, 'Query syntax error')
    if not attr in attrs:
         raise Exception(16, attr, 'Unsupported index')
    import corplib
    if exact_match:
        wlpattern = '^' + value + '$'
    else:
        wlpattern = '.*' + value + '.*'
    wl = corplib.wordlist(corp, wlattr=attr, wlpat=wlpattern, wlsort='f')
    return [(d['str'], d['freq']) for d in wl][start:][:max_ter]

def translate_kwic(corpus1, corpus2, data):
    "translates tokens (tokenums) into lemmata in corpus2 using dict"
    from translate import Translator
    c1 = manatee.Corpus(corpus1)
    c2 = manatee.Corpus(corpus2)
    try:
        alignstruct = c1.get_conf('ALIGNSTRUCT')
        align = c1.get_struct(alignstruct)
    except Exception, e:
        return ({}, {})
    dictname = '%s-%s' % (c1.get_conf('LANGUAGE'), c2.get_conf('LANGUAGE'))
    translator = Translator(dictname)
    if not translator:
        return ({}, {})
    translations = {}
    toknum2words = {}
    try:
        attr = c1.get_attr('lemma')
    except:
        attr = c1.get_attr('word')
    for item in data.split('\t'):
        toknum, hitlen = item.split(':')
        for i in range(int(hitlen)):
            p = int(toknum)+i
            ltt = attr.id2str(attr.pos2id(p)).lower()
            toknum2words[p] = ltt
            if ltt not in translations.keys():
                translations[ltt] = set([(s, w.lower())
                        for s,w in translator.get_translations(ltt)])
            if align:
                align_n = align.num_at_pos(p)
                conc = manatee.Concordance(c1, '<%s #%d>' %\
                        (alignstruct, align_n), 0, -1)
                conc.sync()
                conc.add_aligned(c2.get_conffile())
                conc.switch_aligned(c2.get_conffile())
                for l, w in zip(
                        kwiclines(conc, 0, 1, '-1:s', '1:s', attrs='lemma'),
                        kwiclines(conc, 0, 1, '-1:s', '1:s', attrs='word')):
                    for ctx in ['Left', 'Kwic', 'Right']:
                        if l[ctx] and w[ctx] and l[ctx][0]['str'] \
                                and w[ctx][0]['str']:
                            for t1, t2 in zip(l[ctx][0]['str'].split(),
                                    w[ctx][0]['str'].split()):
                                if t1.lower() in translations[ltt]:
                                    translations[ltt].add((
                                            translations[ltt][t1.lower()][0],
                                            t2.lower()))#x0.9?
    for k in translations:
        translations[k] = sorted(list(translations[k]), reverse=True)
    return (translations, toknum2words)
